"""Implementation of enumerated types.

This module provides the `Enum` class, which can be used to construct
enumerated types.  Those types are defined by providing an *exhaustive
set or list* of possible, named values for a variable of that type.
Enumerated variables of the same type are usually compared between them
for equality and sometimes for order, but are not usually operated upon.

Enumerated values have an associated *name* and *concrete value*.  Every
name is unique and so are concrete values.  An enumerated variable
always takes the concrete value, not its name.  Usually, the concrete
value is not used directly, and frequently it is entirely irrelevant.
For the same reason, an enumerated variable is not usually compared with
concrete values out of its enumerated type.  For that kind of use,
standard variables and constants are more adequate.

"""


__docformat__ = 'reStructuredText'
"""The format of documentation strings in this module."""


class Enum:
    """Enumerated type.

    Each instance of this class represents an enumerated type. The
    values of the type must be declared
    *exhaustively* and named with
    *strings*, and they might be given explicit
    concrete values, though this is not compulsory. Once the type is
    defined, it can not be modified.

    There are three ways of defining an enumerated type. Each one
    of them corresponds to the type of the only argument in the
    constructor of Enum:

    - *Sequence of names*: each enumerated
      value is named using a string, and its order is determined by
      its position in the sequence; the concrete value is assigned
      automatically::

          >>> boolEnum = Enum(['True', 'False'])

    - *Mapping of names*: each enumerated
      value is named by a string and given an explicit concrete value.
      All of the concrete values must be different, or a
      ValueError will be raised::

          >>> priority = Enum({'red': 20, 'orange': 10, 'green': 0})
          >>> colors = Enum({'red': 1, 'blue': 1})
          Traceback (most recent call last):
          ...
          ValueError: enumerated values contain duplicate concrete values: 1

    - *Enumerated type*: in that case, a copy
      of the original enumerated type is created. Both enumerated
      types are considered equal::

          >>> prio2 = Enum(priority)
          >>> priority == prio2
          True

    Please note that names starting with _ are
    not allowed, since they are reserved for internal usage::

        >>> prio2 = Enum(['_xx'])
        Traceback (most recent call last):
        ...
        ValueError: name of enumerated value can not start with ``_``: '_xx'

    The concrete value of an enumerated value is obtained by
    getting its name as an attribute of the Enum
    instance (see __getattr__()) or as an item (see
    __getitem__()). This allows comparisons between
    enumerated values and assigning them to ordinary Python
    variables::

        >>> redv = priority.red
        >>> redv == priority['red']
        True
        >>> redv > priority.green
        True
        >>> priority.red == priority.orange
        False

    The name of the enumerated value corresponding to a concrete
    value can also be obtained by using the
    __call__() method of the enumerated type. In this
    way you get the symbolic name to use it later with
    __getitem__()::

        >>> priority(redv)
        'red'
        >>> priority.red == priority[priority(priority.red)]
        True

    (If you ask, the __getitem__() method is
    not used for this purpose to avoid ambiguity in the case of using
    strings as concrete values.)

    """

    def __init__(self, enum):
        mydict = self.__dict__

        mydict['_names'] = {}
        mydict['_values'] = {}

        if isinstance(enum, list) or isinstance(enum, tuple):
            for (value, name) in enumerate(enum):  # values become 0, 1, 2...
                self._check_and_set_pair(name, value)
        elif isinstance(enum, dict):
            for (name, value) in enum.items():
                self._check_and_set_pair(name, value)
        elif isinstance(enum, Enum):
            for (name, value) in enum._names.items():
                self._check_and_set_pair(name, value)
        else:
            raise TypeError("""\
enumerations can only be created from \
sequences, mappings and other enumerations""")

    def _check_and_set_pair(self, name, value):
        """Check validity of enumerated value and insert it into type."""

        names = self._names
        values = self._values

        if not isinstance(name, str):
            raise TypeError(
                f"name of enumerated value is not a string: {name!r}")
        if name.startswith('_'):
            raise ValueError(
                "name of enumerated value can not start with ``_``: %r"
                % name)
        # This check is only necessary with a sequence base object.
        if name in names:
            raise ValueError(
                "enumerated values contain duplicate names: %r" % name)
        # This check is only necessary with a mapping base object.
        if value in values:
            raise ValueError(
                "enumerated values contain duplicate concrete values: %r"
                % value)

        names[name] = value
        values[value] = name
        self.__dict__[name] = value

    def __getitem__(self, name):
        """Get the concrete value of the enumerated value with that name.

        The name of the enumerated value must be a string. If there is no value
        with that name in the enumeration, a KeyError is raised.

        Examples
        --------

        Let ``enum`` be an enumerated type defined as:

        >>> enum = Enum({'T0': 0, 'T1': 2, 'T2': 5})

        then:

        >>> enum['T1']
        2
        >>> enum['foo']
        Traceback (most recent call last):
          ...
        KeyError: "no enumerated value with that name: 'foo'"

        """

        try:
            return self._names[name]
        except KeyError:
            raise KeyError(f"no enumerated value with that name: {name!r}")

    def __setitem__(self, name, value):
        """This operation is forbidden."""
        raise IndexError("operation not allowed")

    def __delitem__(self, name):
        """This operation is forbidden."""
        raise IndexError("operation not allowed")

    def __getattr__(self, name):
        """Get the concrete value of the enumerated value with that name.

        The name of the enumerated value must be a string. If there is no value
        with that name in the enumeration, an AttributeError is raised.

        Examples
        --------
        Let ``enum`` be an enumerated type defined as:

        >>> enum = Enum({'T0': 0, 'T1': 2, 'T2': 5})

        then:

        >>> enum.T1
        2
        >>> enum.foo
        Traceback (most recent call last):
          ...
        AttributeError: no enumerated value with that name: 'foo'

        """

        try:
            return self[name]
        except KeyError as ke:
            raise AttributeError(*ke.args)

    def __setattr__(self, name, value):
        """This operation is forbidden."""
        raise AttributeError("operation not allowed")

    def __delattr__(self, name):
        """This operation is forbidden."""
        raise AttributeError("operation not allowed")

    def __contains__(self, name):
        """Is there an enumerated value with that name in the type?

        If the enumerated type has an enumerated value with that name, True is
        returned.  Otherwise, False is returned. The name must be a string.

        This method does *not* check for concrete values matching a value in an
        enumerated type. For that, please use the :meth:`Enum.__call__` method.

        Examples
        --------
        Let ``enum`` be an enumerated type defined as:

        >>> enum = Enum({'T0': 0, 'T1': 2, 'T2': 5})

        then:

        >>> 'T1' in enum
        True
        >>> 'foo' in enum
        False
        >>> 0 in enum
        Traceback (most recent call last):
          ...
        TypeError: name of enumerated value is not a string: 0
        >>> enum.T1 in enum  # Be careful with this!
        Traceback (most recent call last):
          ...
        TypeError: name of enumerated value is not a string: 2

        """

        if not isinstance(name, str):
            raise TypeError(
                f"name of enumerated value is not a string: {name!r}")
        return name in self._names

    def __call__(self, value, *default):
        """Get the name of the enumerated value with that concrete value.

        If there is no value with that concrete value in the enumeration and a
        second argument is given as a default, this is returned. Else, a
        ValueError is raised.

        This method can be used for checking that a concrete value belongs to
        the set of concrete values in an enumerated type.

        Examples
        --------
        Let ``enum`` be an enumerated type defined as:

        >>> enum = Enum({'T0': 0, 'T1': 2, 'T2': 5})

        then:

        >>> enum(5)
        'T2'
        >>> enum(42, None) is None
        True
        >>> enum(42)
        Traceback (most recent call last):
          ...
        ValueError: no enumerated value with that concrete value: 42

        """

        try:
            return self._values[value]
        except KeyError:
            if len(default) > 0:
                return default[0]
            raise ValueError(
                f"no enumerated value with that concrete value: {value!r}")

    def __len__(self):
        """Return the number of enumerated values in the enumerated type.

        Examples
        --------
        >>> len(Enum(['e%d' % i for i in range(10)]))
        10

        """

        return len(self._names)

    def __iter__(self):
        """Iterate over the enumerated values.

        Enumerated values are returned as (name, value) pairs *in no particular
        order*.

        Examples
        --------
        >>> enumvals = {'red': 4, 'green': 2, 'blue': 1}
        >>> enum = Enum(enumvals)
        >>> enumdict = dict([(name, value) for (name, value) in enum])
        >>> enumvals == enumdict
        True

        """

        yield from self._names.items()

    def __eq__(self, other):
        """Is the other enumerated type equivalent to this one?

        Two enumerated types are equivalent if they have exactly the same
        enumerated values (i.e. with the same names and concrete values).

        Examples
        --------

        Let ``enum*`` be enumerated types defined as:

        >>> enum1 = Enum({'T0': 0, 'T1': 2})
        >>> enum2 = Enum(enum1)
        >>> enum3 = Enum({'T1': 2, 'T0': 0})
        >>> enum4 = Enum({'T0': 0, 'T1': 2, 'T2': 5})
        >>> enum5 = Enum({'T0': 0})
        >>> enum6 = Enum({'T0': 10, 'T1': 20})

        then:

        >>> enum1 == enum1
        True
        >>> enum1 == enum2 == enum3
        True
        >>> enum1 == enum4
        False
        >>> enum5 == enum1
        False
        >>> enum1 == enum6
        False

        Comparing enumerated types with other kinds of objects produces
        a false result:

        >>> enum1 == {'T0': 0, 'T1': 2}
        False
        >>> enum1 == ['T0', 'T1']
        False
        >>> enum1 == 2
        False

        """

        if not isinstance(other, Enum):
            return False
        return self._names == other._names

    def __ne__(self, other):
        """Is the `other` enumerated type different from this one?

        Two enumerated types are different if they don't have exactly
        the same enumerated values (i.e. with the same names and
        concrete values).

        Examples
        --------

        Let ``enum*`` be enumerated types defined as:

        >>> enum1 = Enum({'T0': 0, 'T1': 2})
        >>> enum2 = Enum(enum1)
        >>> enum3 = Enum({'T1': 2, 'T0': 0})
        >>> enum4 = Enum({'T0': 0, 'T1': 2, 'T2': 5})
        >>> enum5 = Enum({'T0': 0})
        >>> enum6 = Enum({'T0': 10, 'T1': 20})

        then:

        >>> enum1 != enum1
        False
        >>> enum1 != enum2 != enum3
        False
        >>> enum1 != enum4
        True
        >>> enum5 != enum1
        True
        >>> enum1 != enum6
        True

        """

        return not self.__eq__(other)

    # XXX: API incompatible change for PyTables 3 line
    # Overriding __eq__ blocks inheritance of __hash__ in 3.x
    # def __hash__(self):
    #    return hash((self.__class__, tuple(self._names.items())))
    def __repr__(self):
        """Return the canonical string representation of the enumeration. The
        output of this method can be evaluated to give a new enumeration object
        that will compare equal to this one.

        Examples
        --------
        >>> repr(Enum({'name': 10}))
        "Enum({'name': 10})"

        """

        return 'Enum(%s)' % self._names


def _test():
    import doctest
    return doctest.testmod()


if __name__ == '__main__':
    _test()



########################################################################
#
# License: BSD
# Created: November 25, 2009
# Author:  Francesc Alted - faltet@pytables.com
#
# $Id$
#
########################################################################

"""Cython functions and classes for supporting links in HDF5."""

from .exceptions import HDF5ExtError

from .hdf5extension cimport Node
from .utilsextension cimport cstr_to_pystr

from libc.stdlib cimport malloc, free
from libc.string cimport strlen
from cpython.unicode cimport PyUnicode_DecodeUTF8
from .definitions cimport (H5P_DEFAULT,
    hid_t, herr_t, hbool_t, int64_t, H5T_cset_t, haddr_t)



#----------------------------------------------------------------------

# External declarations

cdef extern from "H5Lpublic.h" nogil:

  ctypedef enum H5L_type_t:
    H5L_TYPE_ERROR = (-1),       # Invalid link type id
    H5L_TYPE_HARD = 0,           # Hard link id
    H5L_TYPE_SOFT = 1,           # Soft link id
    H5L_TYPE_EXTERNAL = 64,      # External link id
    H5L_TYPE_MAX = 255           # Maximum link type id

  # Information struct for link (for H5Lget_info)
  cdef union _add_u:
    haddr_t address              # Address hard link points to
    size_t val_size              # Size of a soft link or UD link value

  ctypedef struct H5L_info_t:
    H5L_type_t     type          # Type of link
    hbool_t        corder_valid  # Indicate if creation order is valid
    int64_t        corder        # Creation order
    H5T_cset_t     cset          # Character set of link name
    _add_u         u             # Size of a soft link or UD link value

  # Operations with links
  herr_t H5Lcreate_hard(
    hid_t obj_loc_id, char *obj_name, hid_t link_loc_id, char *link_name,
    hid_t lcpl_id, hid_t lapl_id)

  herr_t H5Lcreate_soft(
    char *target_path, hid_t link_loc_id, char *link_name,
    hid_t lcpl_id, hid_t lapl_id)

  herr_t H5Lcreate_external(
    char *file_name, char *object_name, hid_t link_loc_id, char *link_name,
    hid_t lcpl_id, hid_t lapl_id)

  herr_t H5Lget_info(
    hid_t link_loc_id, char *link_name, H5L_info_t *link_buff,
    hid_t lapl_id)

  herr_t H5Lget_val(
    hid_t link_loc_id, char *link_name, void *linkval_buff, size_t size,
    hid_t lapl_id)

  herr_t H5Lunpack_elink_val(
    char *ext_linkval, size_t link_size, unsigned *flags,
    const char **filename, const char **obj_path)

  herr_t H5Lcopy(
    hid_t src_loc_id, char *src_name, hid_t dest_loc_id, char *dest_name,
    hid_t lcpl_id, hid_t lapl_id)


#----------------------------------------------------------------------

# Helper functions

def _get_link_class(parent_id, name):
    """Guess the link class."""

    cdef herr_t ret
    cdef H5L_info_t link_buff
    cdef H5L_type_t link_type

    ret = H5Lget_info(parent_id, name, &link_buff, H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to get info about link")

    link_type = link_buff.type
    if link_type == H5L_TYPE_SOFT:
      return "SoftLink"
    elif link_type == H5L_TYPE_EXTERNAL:
      return "ExternalLink"
    # elif link_type == H5L_TYPE_HARD:
    #   return "HardLink"
    else:
      return "UnImplemented"




def _g_create_hard_link(parentnode, str name, targetnode):
  """Create a hard link in the file."""

  cdef herr_t ret
  cdef bytes encoded_name = name.encode('utf-8')
  cdef bytes encoded_v_name = targetnode._v_name.encode('utf-8')

  ret = H5Lcreate_hard(targetnode._v_parent._v_objectid, encoded_v_name,
                       parentnode._v_objectid, <char*>encoded_name,
                       H5P_DEFAULT, H5P_DEFAULT)
  if ret < 0:
    raise HDF5ExtError("failed to create HDF5 hard link")




#----------------------------------------------------------------------

# Public classes

cdef class Link(Node):
  """Extension class from which all link extensions inherits."""

  def _g_copy(self, newparent, newname, recursive, _log=True, **kwargs):
    """Private part for the _f_copy() method."""

    cdef herr_t ret
    cdef object stats
    cdef bytes encoded_name, encoded_newname

    encoded_name = self.name.encode('utf-8')
    encoded_newname = newname.encode('utf-8')

    # @TODO: set property list --> utf-8
    ret = H5Lcopy(self.parent_id, encoded_name, newparent._v_objectid,
                  encoded_newname, H5P_DEFAULT, H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to copy HDF5 link")

    # Update statistics if needed.
    stats = kwargs.get('stats', None)
    if stats is not None:
      stats['links'] += 1

    return newparent._v_file.get_node(newparent, newname)


cdef class SoftLink(Link):
  """Extension class representing a soft link."""

  def _g_create(self):
    """Create the link in file."""

    cdef herr_t ret
    cdef bytes encoded_name = self.name.encode('utf-8')
    cdef bytes encoded_target = self.target.encode('utf-8')

    ret = H5Lcreate_soft(encoded_target, self.parent_id, encoded_name,
                         H5P_DEFAULT, H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to create HDF5 soft link")

    return 0  # Object ID is zero'ed, as HDF5 does not assign one for links

  def _g_open(self):
    """Open the link in file."""

    cdef herr_t ret
    cdef H5L_info_t link_buff
    cdef size_t val_size
    cdef char *clinkval
    cdef bytes encoded_name

    encoded_name = self.name.encode('utf-8')

    ret = H5Lget_info(self.parent_id, encoded_name, &link_buff, H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to get info about soft link")

    val_size = link_buff.u.val_size
    clinkval = <char *>malloc(val_size)

    ret = H5Lget_val(self.parent_id, encoded_name, clinkval, val_size,
                     H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to get target value")

    self.target = PyUnicode_DecodeUTF8(clinkval, strlen(clinkval), NULL)

    # Release resources
    free(clinkval)
    return 0  # Object ID is zero'ed, as HDF5 does not assign one for links


cdef class ExternalLink(Link):
  """Extension class representing an external link."""

  def _g_create(self):
    """Create the link in file."""

    cdef herr_t ret
    cdef bytes encoded_name, encoded_filename, encoded_target

    encoded_name = self.name.encode('utf-8')

    filename, target = self._get_filename_node()
    encoded_filename = filename.encode('utf-8')
    encoded_target = target.encode('utf-8')

    ret = H5Lcreate_external(<char*>encoded_filename, <char*>encoded_target,
                             self.parent_id, <char*>encoded_name,
                             H5P_DEFAULT, H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to create HDF5 external link")

    return 0  # Object ID is zero'ed, as HDF5 does not assign one for links

  def _g_open(self):
    """Open the link in file."""

    cdef herr_t ret
    cdef H5L_info_t link_buff
    cdef size_t val_size
    cdef char *clinkval
    cdef char *cfilename
    cdef char *c_obj_path
    cdef unsigned flags
    cdef bytes encoded_name
    cdef str filename, obj_path

    encoded_name = self.name.encode('utf-8')

    ret = H5Lget_info(self.parent_id, encoded_name, &link_buff, H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to get info about external link")

    val_size = link_buff.u.val_size
    clinkval = <char *>malloc(val_size)

    ret = H5Lget_val(self.parent_id, encoded_name, clinkval, val_size,
                     H5P_DEFAULT)
    if ret < 0:
      raise HDF5ExtError("failed to get target value")

    ret = H5Lunpack_elink_val(clinkval, val_size, &flags,
                              <const char **>&cfilename,
                              <const char **>&c_obj_path)
    if ret < 0:
      raise HDF5ExtError("failed to unpack external link value")

    filename = cstr_to_pystr(cfilename)
    obj_path = cstr_to_pystr(c_obj_path)

    self.target = filename+':'+obj_path

    # Release resources
    free(clinkval)
    return 0  # Object ID is zero'ed, as HDF5 does not assign one for links


## Local Variables:
## mode: python
## py-indent-offset: 2
## tab-width: 2
## fill-column: 78
## End:
